/*	pbpac.h

	Applicative Cache layer for PBP.

	See pbp.h for copyright, etc.
*/

typedef	struct {
	AoBreg_t op;		// opcode
	AoBreg_t a, b, c;	// c = a op b
}	AC_t;

#define	ACSIZE	(32 * 1024)	// size of applicative cache

class AC
{
	// Applicative Cache
	AC_t ac[ACSIZE];
	int spot;
	int hit, miss, rehash;

public:
	AC() {
		bzero(((void *) &ac[0]), sizeof(ac));
		hit = 0;
		miss = 0;
		rehash = 0;
	}

	int hash(AoBreg_t op, AoBreg_t a, AoBreg_t b) {
		// should really pick this by benchmarking...
		register int h = (op * (ACSIZE/256));
		h ^= a;
		h ^= b * (ACSIZE/AOBREGS);
		return(h & (ACSIZE-1));
	}

	int avail(AoBreg_t op, AoBreg_t a, AoBreg_t b) {
		register int h = hash(op, a, b);

		for (;;) {
			// empty entry
			if (!ac[h].op) {
				++miss;
				ac[h].op = op;
				ac[h].a = a;
				ac[h].b = b;
				spot = h;
				return(-1);
			}

			// found it
			if ((ac[h].op == op) &&
			    (ac[h].a == a) &&
			    (ac[h].b == b)) {
				++hit;
				return(ac[h].c);
			}

			++rehash;
			h = ((h + 13) & (ACSIZE-1));
		}
	}

	int assoc(AoBreg_t op, AoBreg_t a, AoBreg_t b) {
		// assoiciative operator, so normalize operands
		return((b < a) ? avail(op, a, b) : avail(op, b, a));
	}

	void Stats() {
		fprintf(stderr, "AC: %d hits, %d misses, %d rehashes\n", hit, miss, rehash);
	}

	AoBreg_t cache(AoBreg_t c) {
		return(ac[spot].c = c);
	}
};
